/*******************************************************************\

Module: Verilog Synthesis

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#include <assert.h>

#include <map>
#include <set>

#include <arith_tools.h>
#include <location.h>
#include <expr_util.h>
#include <simplify_expr.h>
#include <bitvector.h>
#include <identifier.h>

#include "expr2verilog.h"
#include "verilog_synthesis.h"
#include "verilog_expr.h"

/*******************************************************************\

Function: verilog_synthesist::synth_expr

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_expr(
  exprt &expr,
  symbol_statet symbol_state)
{
  if(expr.id()==ID_symbol)
  {
    const symbolt &symbol=lookup(expr.get(ID_identifier));
    
    // does it have a value?
    valuest::const_iterator v_it=values.find(symbol.name);

    if(v_it!=values.end())
    {
      exprt c=from_integer(v_it->second, integer_typet());
      c.location()=expr.location();
      expr=c;
    }
    else if(symbol.is_macro)
    {
      // substitute
      assert(symbol.value.is_not_nil());
      expr=symbol.value;
      
      // recursive call
      synth_expr(expr, symbol_state);
    }
    else
    {
      switch(symbol_state)
      {
      case S_SYMBOL:
        break;

      case S_CURRENT:
        current_value(symbol, expr);
        break;

      case S_FINAL:
        final_value(symbol, expr);
        break;

      case S_NONE:
        err_location(expr);
        throw "symbols not allowed here";
      }
    }
  }
  else if(expr.id()=="functioncall")
  {
    expand_functioncall(to_functioncall_expr(expr));
  }
  else if(expr.id()==ID_hierarchical_identifier)
  {
    expand_hierarchical_identifier(
      to_hierarchical_identifier_expr(expr),
      symbol_state);
  }
  else if(expr.id()==ID_typecast)
  {
    assert(expr.operands().size()==1);
    synth_expr(expr.op0(), symbol_state);

    // we perform some form of simplification for these
    if(expr.op0().is_constant())
      simplify(expr, ns);
  }
  else if(expr.has_operands())
    Forall_operands(it, expr)
      synth_expr(*it, symbol_state);   
}

/*******************************************************************\

Function: verilog_synthesist::expand_functioncall

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::expand_functioncall(functioncall_exprt &call)
{
  // check some restrictions
  if(construct!=C_INITIAL &&
     construct!=C_ALWAYS)
  {
    err_location(call);
    throw "function call not allowed here";
  }

  // this is essentially inlined
  const symbol_exprt &function=to_symbol_expr(call.function());
  
  const symbolt &symbol=lookup(function.get_identifier());
  
  if(symbol.type.id()!=ID_code)
  {
    err_location(call);
    throw "function call expects function argument";
  }
  
  const code_typet &code_type=to_code_type(symbol.type);

  if(code_type.return_type()==empty_typet())
  {
    err_location(call);
    throw "function call cannot call task";
  }
  
  const code_typet::argumentst &arguments=
    code_type.arguments();
    
  const exprt::operandst &actuals=
    call.op1().operands();
    
  if(arguments.size()!=actuals.size())
  {
    err_location(call);
    throw "wrong number of arguments";
  }
  
  // do assignments to function arguments
  for(unsigned i=0; i<arguments.size(); i++)
  {
    const symbolt &a_symbol=lookup(arguments[i].get_identifier());
    verilog_blocking_assignt assignment;
    assignment.lhs()=::symbol_expr(a_symbol);
    assignment.rhs()=actuals[i];
    assignment.location()=call.location();
    synth_statement(assignment);
  }

  synth_statement(to_verilog_statement(symbol.value));
  
  // replace function call by return value symbol
  const symbolt &return_symbol=
    lookup(id2string(symbol.name)+"."+
           id2string(symbol.base_name));

  exprt return_value=::symbol_expr(return_symbol);

  call.swap(return_value);
}

/*******************************************************************\

Function: verilog_synthesist::expand_hierarchical_identifier

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::expand_hierarchical_identifier(
  hierarchical_identifier_exprt &expr,
  symbol_statet symbol_state)
{
  synth_expr(expr.op0(), symbol_state);

  if(expr.op0().id()!=ID_symbol)
  {
    err_location(expr);
    throw "synthesis expected symbol on lhs of `.'";
  }

  if(expr.op0().type().id()!=ID_module_instance)
  {
    err_location(expr);
    str << "synthesis expected module instance on lhs of `.', but got `"
        << to_string(expr.op0().type()) << "'";
    throw 0;
  }
  
  const irep_idt &lhs_identifier=expr.op0().get(ID_identifier);

  // rhs

  if(expr.op1().id()!=ID_symbol)
  {
    err_location(expr);
    str << "synthesis expected symbol on rhs of `.', but got `"
        << to_string(expr.op1()) << "'";
    throw 0;
  }

  const irep_idt &rhs_identifier=expr.op1().get(ID_identifier);

  // just patch together
  
  std::string full_identifier=
    id2string(lhs_identifier)+"."+id2string(rhs_identifier);

  // Note: may not yet be in symbol table, as the inst module
  // item may be later.
  // The type checker already checked that it's fine.

  symbol_exprt new_symbol;
  new_symbol.type()=expr.type();
  new_symbol.location()=expr.location();
  new_symbol.set_identifier(full_identifier);
  expr.swap(new_symbol);
}

/*******************************************************************\

Function: verilog_synthesist::assignment

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::assignment(
  const exprt &lhs,
  const exprt &rhs,
  bool blocking)
{
  if(lhs.id()==ID_concatenation) // split it up
                                // TODO: split it up more intelligently
  {
    unsigned offset=0;
    
    // do it from right to left
    
    for(exprt::operandst::const_reverse_iterator
        it=lhs.operands().rbegin();
        it!=lhs.operands().rend();
        it++)
    {
      exprt offset_constant(ID_constant, typet(ID_natural));
      offset_constant.set(ID_value, offset);

      if(it->type().id()==ID_bool)
      {
        exprt bit_extract(ID_extractbit, it->type());
        bit_extract.copy_to_operands(rhs, offset_constant);
        offset++;

        assignment(*it, bit_extract, blocking);
      }
      else if(it->type().id()==ID_signedbv ||
              it->type().id()==ID_unsignedbv)
      {
        unsigned width=get_width(it->type());
      
        exprt offset_constant2(ID_constant, typet(ID_natural));
        offset_constant2.set(ID_value, offset+width-1);

        exprt bit_extract(ID_extractbits, it->type());
        bit_extract.copy_to_operands(rhs, offset_constant, offset_constant2);
        
        assignment(*it, bit_extract, blocking);
        
        offset+=width;
      }
      else
      {
        err_location(*it);
        str << "assignment to type `" << to_string(it->type())
            << "' not supported";
      }
    }
    
    return;
  }
 
  // get identifier
  const symbolt &symbol=assignment_symbol(lhs);

  if(!symbol.is_statevar)
  {
    err_location(lhs);
    throw "assignment to non-register";
  }

  if(construct==C_ALWAYS && event_guard==G_NONE)
  {
    err_location(lhs);
    throw "permanent assignment without event guard";
  }

  {
    event_guardt new_type;
    if(construct==C_ALWAYS)  
      new_type=event_guard;
    else
      new_type=G_CLOCK;

    assignmentt &assignment=assignments[symbol.name];

    if(assignment.type==G_NONE)
      assignment.type=new_type;
    else if(assignment.type!=new_type)
    {
      err_location(lhs);
      str << "conflicting assignment types (new: "
          << as_string(new_type) << ", old: "
          << as_string(assignment.type) << ")";
      throw 0;      
    }
    
    membert member;
    assignment_member_rec(lhs, member,
      (construct==C_INITIAL)?assignment.init:assignment.next);
  }

  {
    assert(value_map!=NULL);

    exprt new_rhs(rhs), new_value;
    assignment_rec(lhs, new_rhs, new_value);

    if(blocking)
      value_map->current.assign(symbol.name, new_value);

    value_map->final.assign(symbol.name, new_value);
  }
}

/*******************************************************************\

Function: verilog_synthesist::assignment_rec

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::assignment_rec(
  const exprt &lhs,
  exprt &rhs,
  exprt &new_value)
{
  if(lhs.id()==ID_symbol)
  {
    new_value.swap(rhs);
    rhs.clear();
  }
  else if(lhs.id()==ID_index ||
          lhs.id()==ID_extractbit)
  {
    if(lhs.operands().size()!=2)
      throw "index takes two operands";

    const exprt &lhs_array = lhs.op0();
    const exprt &lhs_index = lhs.op1();

    // turn
    //   a[i]=e
    // into
    //   a'==a WITH [i:=e]

    exprt new_rhs(ID_with, lhs_array.type());

    new_rhs.reserve_operands(3);
    new_rhs.copy_to_operands(lhs_array);
    new_rhs.copy_to_operands(lhs_index);
    new_rhs.move_to_operands(rhs);

    // do the array
    synth_expr(new_rhs.op0(), S_FINAL);

    // do the index
    synth_expr(new_rhs.op1(), S_CURRENT);

    // do the value
    assignment_rec(lhs_array, new_rhs, new_value);
  }
  else if(lhs.id()==ID_extractbits)
  {
    // we flatten n-bit extractbits into n times extractbit

    if(lhs.operands().size()!=3)
      throw "extractbits takes three operands";

    const exprt &lhs_bv        = lhs.op0();
    const exprt &lhs_index_one = lhs.op1();
    const exprt &lhs_index_two = lhs.op2();

    mp_integer from, to;

    if(to_integer(lhs_index_one, from))
    {
      err_location(lhs_index_one);
      throw "failed to convert range";
    }

    if(to_integer(lhs_index_two, to))
    {
      err_location(lhs_index_two);
      throw "failed to convert range";
    }

    if(from>to)
      std::swap(from, to);

    // redundant?
    if(from==0 &&
       to==get_width(lhs_bv.type())-1)
    {
      assignment_rec(lhs_bv, rhs, new_value);
      return;
    }

    // turn
    //   a[i]=e
    // into
    //   a'==a WITH [i:=e]

    exprt synth_lhs_bv(lhs_bv);

    // do the array, but just once
    synth_expr(synth_lhs_bv, S_FINAL);

    exprt last_value;
    last_value.make_nil();

    // now add the indexes in the range
    for(mp_integer i=from; i<=to; ++i)
    {
      exprt offset=from_integer(i-from, integer_typet());

      exprt rhs_extractbit(ID_extractbit, typet(ID_bool));
      rhs_extractbit.reserve_operands(2);
      rhs_extractbit.copy_to_operands(rhs);
      rhs_extractbit.move_to_operands(offset);

      exprt count=from_integer(i, integer_typet());

      exprt new_rhs(ID_with, lhs_bv.type());
      new_rhs.reserve_operands(3);
      new_rhs.copy_to_operands(synth_lhs_bv);
      new_rhs.move_to_operands(count);
      new_rhs.move_to_operands(rhs_extractbit);

      // do the value
      assignment_rec(lhs_bv, new_rhs, new_value);

      if(last_value.is_nil())
        last_value.swap(new_value);
      else
      {
        // merge the withs
        assert(new_value.id()==ID_with);
        assert(new_value.operands().size()==3);
        assert(last_value.id()==ID_with);
        assert(last_value.operands().size()>=3);

        last_value.move_to_operands(new_value.op1());
        last_value.move_to_operands(new_value.op2());
      }
    }

    new_value.swap(last_value);
  }
  else
    throw "unexpected lhs: "+lhs.id_string();

  #if 0
  // do "with" merging

  if(new_value.id()==ID_with &&
     new_value.op0().id()==ID_with)
  {
    exprt tmp;

    tmp.swap(new_value.op0());

    tmp.reserve_operands(tmp.operands().size()+2);
    tmp.move_to_operands(new_value.op1());
    tmp.move_to_operands(new_value.op2());

    new_value.swap(tmp);
  }
  #endif
}

/*******************************************************************\

Function: verilog_synthesist::assignment_member_rec

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::assignment_member_rec(
  const exprt &lhs,
  membert &member,
  assignmentt::datat &data)
{
  if(lhs.id()==ID_symbol)
  {
    // done
    add_assignment_member(lhs, member, data);
  }
  else if(lhs.id()==ID_index ||
          lhs.id()==ID_extractbit)
  {
    if(lhs.operands().size()!=2)
      throw "index takes two operands";
      
    // constant index?
    mp_integer index;
    if(to_integer(lhs.op1(), index))
    {
      // done
      add_assignment_member(lhs, member, data);
    }
    else
    {
      // do the value
      member.push_back(index);
      assignment_member_rec(lhs.op0(), member, data);
      member.pop_back();
    }
  }
  else if(lhs.id()==ID_extractbits)
  {
    // we flatten n-bit extractbits into n times extractbit

    if(lhs.operands().size()!=3)
      throw "extractbits takes three operands";

    const exprt &lhs_bv        = lhs.op0();
    const exprt &lhs_index_one = lhs.op1();
    const exprt &lhs_index_two = lhs.op2();

    mp_integer from, to;

    if(to_integer(lhs_index_one, from))
    {
      err_location(lhs_index_one);
      throw "failed to convert range";
    }

    if(to_integer(lhs_index_two, to))
    {
      err_location(lhs_index_two);
      throw "failed to convert range";
    }

    if(from>to)
      std::swap(from, to);

    member.push_back(mp_integer());

    // now add the indexes in the range
    for(mp_integer i=from; i<=to; ++i)
    {
      // do the value
      member.back()=i;
      assignment_member_rec(lhs_bv, member, data);
    }
    
    member.pop_back();
  }
  else
    throw "unexpected lhs: "+lhs.id_string();
}

/*******************************************************************\

Function: verilog_synthesist::disjoint

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool verilog_synthesist::disjoint(
  const membert &a,
  const membert &b)
{
  membert::const_iterator a_it=a.begin();
  membert::const_iterator b_it=b.begin();
  
  while(a_it!=a.end() && b_it!=b.end())
  {
    if(*a_it!=*b_it) return true;
    a_it++, b_it++;
  }
  
  return false;
}

/*******************************************************************\

Function: verilog_synthesist::add_assignment_member

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::add_assignment_member(
  const exprt &lhs,
  const membert &member,
  assignmentt::datat &data)
{
  // see if we conflict with any previous assignment
  for(member_listt::const_iterator
      it=data.assigned_previous.begin();
      it!=data.assigned_previous.end();
      it++)
  {
    if(!disjoint(member, *it))
    {
      err_location(lhs);
      str << "conflict with previous assignment";
      throw 0;
    }
  }

  data.assigned_current.push_back(member);
}

/*******************************************************************\

Function: verilog_synthesist::assignment_symbol

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

symbolt &verilog_synthesist::assignment_symbol(const exprt &lhs)
{
  const exprt *e=&lhs;

  while(1)
  {
    if(e->id()==ID_index)
    {
      if(e->operands().size()!=2)
        throw "index takes two operands";

      e=&e->op0();
    }
    else if(e->id()==ID_extractbit)
    {
      if(e->operands().size()!=2)
        throw "extractbit takes two operands";

      e=&e->op0();
    }
    else if(e->id()==ID_extractbits)
    {
      if(e->operands().size()!=3)
        throw "extractbits takes three operands";

      e=&e->op0();
    }
    else if(e->id()==ID_symbol)
    {
      // get identifier

      const irep_idt &identifier=e->get(ID_identifier);

      contextt::symbolst::iterator it=
        context.symbols.find(identifier);

      if(it==context.symbols.end())
      {
        str << "assignment failed to find symbol `"
            << identifier
            << "'";
        throw 0;
      }

      return it->second;
    }
    else
    {
      str << *e << std::endl;
      str << "synthesis: failed to get identifier";
      error();
      throw 0;
    }
  }
}

/*******************************************************************\

Function: verilog_synthesist::replace_symbols

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool verilog_synthesist::replace_symbols(
  const replace_mapt &what,
  exprt &dest)
{
  bool result=true;

  if(dest.id()==ID_next_symbol ||
     dest.id()==ID_symbol)
  {
    replace_mapt::const_iterator it=
      what.find(dest.get(ID_identifier));

    if(it!=what.end())
    {
      bool is_next_symbol=dest.id()==ID_next_symbol;
      dest=it->second;      

      if(is_next_symbol)
        replace_symbols(ID_next_symbol, dest);

      result=false;
    }
  }
  else
  {
    Forall_operands(it, dest)
      result=replace_symbols(what, *it) && result;
  }

  return result;
}

/*******************************************************************\

Function: verilog_synthesist::replace_symbols

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::replace_symbols(
  const irep_idt &target,
  exprt &dest)
{
  if(dest.id()==ID_symbol)
    dest.id(target);
  else
    Forall_operands(it, dest)
      replace_symbols(target, *it);
}

/*******************************************************************\

Function: verilog_synthesist::instantiate_port

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::instantiate_port(
  const exprt &symbol_expr,
  const exprt &value,
  const replace_mapt &replace_map,
  transt &trans)
{
  replace_mapt::const_iterator it=
    replace_map.find(symbol_expr.get(ID_identifier));

  if(it==replace_map.end())
  {
    str << "failed to find " << symbol_expr << " in replace_map"
        << std::endl;
    throw 0;
  }

  equality_exprt equality(it->second, value);

  if(equality.op0().type()!=equality.op1().type())
    equality.op0().make_typecast(equality.op1().type());

  trans.invar().move_to_operands(equality);
}

/*******************************************************************\

Function: verilog_synthesist::instantiate_ports

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::instantiate_ports(
  const irep_idt &instance,
  const exprt &inst,
  const symbolt &symbol,
  const replace_mapt &replace_map,
  transt &trans)
{
  const irept::subt &ports=symbol.type.find(ID_ports).get_sub();

  if(inst.operands().size()!=ports.size())
  {
    err_location(inst);
    str << "wrong number of ports: expected " << ports.size() 
        << " but got " << inst.operands().size() << std::endl; 
    str << inst.op0() << std::endl;
    throw 0;
  }

  if(inst.operands().size()==0)
    return;

  // named port connection?

  if(inst.op0().id()==ID_named_port_connection)
  {
    forall_operands(o_it, inst)
    {
      if(o_it->operands().size()==2)
      {
        const exprt &op0=o_it->op0();
        const exprt &op1=o_it->op1();

        if(op1.is_not_nil())
          instantiate_port(op0, op1, replace_map, trans);
      }
    }
  }
  else // just a list without names
  {
    irept::subt::const_iterator p_it=
      ports.begin();

    forall_operands(o_it, inst)
    {
      instantiate_port((exprt &)(*p_it), *o_it,
                       replace_map, trans);
      p_it++;
    }
  }
}

/*******************************************************************\

Function: verilog_synthesist::synth_inst

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_inst(
  const verilog_instt &statement,
  transt &trans)
{
  const irep_idt &identifier=statement.get(ID_module);

  // must be in context
  const symbolt &symbol=lookup(identifier);
  
  // make sure it's synthesized already
  verilog_synthesis(
    context, identifier, get_message_handler(), options);

  forall_operands(it, statement)
    expand_inst(symbol, *it, trans);
}

/*******************************************************************\

Function: verilog_synthesist::synth_inst_builtin

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_inst_builtin(
  const verilog_module_itemt &module_item,
  transt &trans)
{
  const irep_idt &module=module_item.get(ID_module);

  forall_operands(it, module_item)
  {
    const exprt &instance=*it;

    // check built-in ones
    if(module==ID_bufif0 ||
       module==ID_bufif1 ||
       module==ID_notif0 ||
       module==ID_notif1)
    {
      // add to general constraints
    
      exprt constraint=*it;
      constraint.id("verilog-primitive-module");
      constraint.type()=typet(ID_bool);
      constraint.set(ID_module, module);
    
      assert(trans.operands().size()==3);
      trans.invar().move_to_operands(constraint);
    }
    else if(module==ID_nmos ||
            module==ID_pmos ||
            module==ID_rnmos ||
            module==ID_rpmos)
    {
      // add to general constraints
    
      exprt constraint=*it;
      constraint.id("verilog-primitive-module");
      constraint.type()=typet(ID_bool);
      constraint.set(ID_module, module);
    
      assert(trans.operands().size()==3);
      trans.invar().move_to_operands(constraint);
    }
    else if(module==ID_and ||
            module==ID_nand ||
            module==ID_or ||
            module==ID_nor ||
            module==ID_xor ||
            module==ID_xnor)
    {
      assert(instance.operands().size()>=2);

      if(instance.operands().size()==2)
      {
        equality_exprt constraint;
        constraint.lhs()=instance.operands()[0];
        constraint.rhs()=instance.operands().back();
        trans.invar().move_to_operands(constraint);
      }
      else
      {
        for(unsigned i=1; i<=instance.operands().size()-2; i++)
        {
          exprt op(module, instance.type());

          if(i==1)
            op.copy_to_operands(instance.operands()[i],
                                instance.operands()[i+1]);
          else
            op.copy_to_operands(instance.operands()[0],
                                instance.operands()[i+1]);

          if(instance.type().id()!=ID_bool)
            op.id("bit"+op.id_string());

          equality_exprt constraint(instance.op0(), op);
          assert(trans.operands().size()==3);
          trans.invar().move_to_operands(constraint);
        }
      }
      
      /*assert(instance.operands().size()!=3);      
      op.move_to_operands(instance.op1(), instance.op2());
      
      if(instance.type().id()!=ID_bool)
        op.id("bit"+op.id_string());
      
      equality_exprt constraint(instance.op0(), op);
    
      assert(trans.operands().size()!=3);
      trans.invar().move_to_operands(constraint);
      */
    }
    else if(module==ID_buf)
    {
      assert(instance.operands().size()>=2);

      for(unsigned i=0; i<instance.operands().size()-1; i++)
      {
        equality_exprt constraint;
        constraint.lhs()=instance.operands()[i];
        constraint.rhs()=instance.operands().back();
    
        assert(trans.operands().size()==3);
        trans.invar().move_to_operands(constraint);
      }
    }
    else if(module==ID_not)
    {
      assert(instance.operands().size()>=2);

      for(unsigned i=0; i<instance.operands().size()-1; i++)
      {
        exprt op(ID_not, instance.type());
        op.copy_to_operands(instance.operands()[i]);
      
        if(instance.type().id()!=ID_bool)
          op.id("bit"+op.id_string());
      
        equality_exprt constraint;
        constraint.lhs()=op;
        constraint.rhs()=instance.operands().back();
    
        assert(trans.operands().size()==3);
        trans.invar().move_to_operands(constraint);
      }
    }
    else if(module=="tranif0" ||
            module=="tranif1" ||
            module=="rtranif1" ||
            module=="rtranif0")
    {
      // add to general constraints
    
      exprt constraint=*it;
      constraint.id("verilog-primitive-module");
      constraint.type()=typet(ID_bool);
      constraint.set(ID_module, module);
    
      assert(trans.operands().size()==3);
      trans.invar().move_to_operands(constraint);
    }
    else if(module=="tran"  ||
            module=="rtran")
    {
      // add to general constraints
    
      exprt constraint=*it;
      constraint.id("verilog-primitive-module");
      constraint.type()=typet(ID_bool);
      constraint.set(ID_module, module);
    
      assert(trans.operands().size()==3);
      trans.invar().move_to_operands(constraint);
    }
    else
    {
      err_location(module_item);
      str << "do not know how to synthesize " << module << std::endl;
      throw 0;
    }
  }
}

/*******************************************************************\

Function: verilog_synthesist::expand_inst

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::expand_inst(
  const symbolt &symbol,
  const exprt &op,
  transt &trans)
{
  construct=C_OTHER;

  const irep_idt &instance=op.get(ID_instance);

  replace_mapt replace_map;

  std::list<irep_idt> new_symbols;

  forall_symbol_module_map(it, context.symbol_module_map, symbol.module)
  {
    const symbolt &symbol=lookup(it->second);

    if(symbol.type.id()!=ID_module)
    {
      // instantiate symbol

      symbolt new_symbol(symbol);

      new_symbol.module=module;
      new_symbol.hierarchy.push_front(id2string(module)+"."+id2string(instance));

      std::string full_identifier;

      full_identifier=mode+"::";
      full_identifier+=id2string(verilog_module_name(module));

      for(symbolt::hierarchyt::const_iterator
          it=new_symbol.hierarchy.begin();
          it!=new_symbol.hierarchy.end();
          it++)
        full_identifier+="."+id2string(lookup(*it).base_name);

      full_identifier+="."+id2string(symbol.base_name);

      new_symbol.name=full_identifier;

      if(context.add(new_symbol))
      {
        str << "name collision during module instantiation: "
            << new_symbol.name << std::endl;
        throw 0;
      }

      new_symbols.push_back(new_symbol.name);

      // build replace map

      std::pair<irep_idt, exprt> replace_pair;
      replace_pair.first=symbol.name;
      symbol_expr(new_symbol, false, replace_pair.second);
      replace_map.insert(replace_pair);
    }
  }

  // replace identifiers in macros

  for(std::list<irep_idt>::const_iterator
      it=new_symbols.begin();
      it!=new_symbols.end();
      it++)
  {
    symbolt &symbol=context_lookup(*it);
    replace_symbols(replace_map, symbol.value);
  }

  // do the trans

  {
    exprt tmp=symbol.value;

    if(tmp.id()!=ID_trans || tmp.operands().size()!=3)
    {
      err_location(op);
      str << "Expected transition system, but got `"
          << tmp.id_string() << "'" << std::endl;
      throw 0;
    }

    replace_symbols(replace_map, tmp);

    for(unsigned i=0; i<3; i++)
      trans.operands()[i].move_to_operands(tmp.operands()[i]);
  }

  instantiate_ports(instance, op, symbol, replace_map, trans);
}

/*******************************************************************\

Function: verilog_synthesist::synth_always

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_always(
  const verilog_alwayst &module_item)
{
  if(module_item.operands().size()!=1)
  {
    err_location(module_item);
    throw "always module item expected to have one operand";
  }

  construct=C_ALWAYS;
  event_guard=G_NONE;

  value_mapt always_value_map;
  value_map=&always_value_map;

  synth_statement(module_item.statement());

  for(std::set<irep_idt>::const_iterator it=
      value_map->final.changed.begin();
      it!=value_map->final.changed.end();
      it++)
  {
    assignmentt &assignment=assignments[*it];
    assignment.next.value=value_map->final.symbol_map[*it];
    assignment.next.move_assignments();
  }

  value_map=NULL;
}

/*******************************************************************\

Function: verilog_synthesist::synth_initial

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_initial(
  const verilog_initialt &module_item)
{
  if(module_item.operands().size()!=1)
  {
    err_location(module_item);
    throw "initial module item expected to have one operand";
  }

  construct=C_INITIAL;
  event_guard=G_NONE;

  value_mapt initial_value_map;
  value_map=&initial_value_map;

  synth_statement(module_item.statement());
  
  for(std::set<irep_idt>::const_iterator it=
      value_map->final.changed.begin();
      it!=value_map->final.changed.end();
      it++)
  {
    assignmentt &assignment=assignments[*it];
    assignment.init.value=value_map->final.symbol_map[*it];
    assignment.init.move_assignments();    
  }

  value_map=NULL;
}

/*******************************************************************\

Function: verilog_synthesist::synth_decl

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_decl(
  const verilog_declt &statement,
  transt &trans)
{
  construct=C_INITIAL;
  event_guard=G_NONE;

  forall_operands(it, statement)
  {
    if(it->id()==ID_equal)
    {
      if(it->operands().size()!=2)
        throw "expected two operands in assignment";

      exprt lhs=it->op0();
      exprt rhs=it->op1();

      if(lhs.id()!=ID_symbol)
      {
        str << "expected symbol on left hand side of assignment"
            << " but got `" << to_string(lhs) << "'";
        throw 0;
      }

      const symbolt &symbol=lookup(lhs.get(ID_identifier));
      assignmentt &assignment=assignments[symbol.name];

      if(symbol.is_statevar)
      {
        synth_expr(rhs, S_CURRENT);

        if(assignment.type==G_NONE)
          assignment.type=G_CLOCK;
        else if(assignment.type!=G_CLOCK)
        {
          err_location(statement);
          throw "variable is combinational";
        }

        if(!assignment.init.value.is_nil())
        {
          err_location(statement);
          throw "variable already has initial value";
        }

        assignment.init.value=rhs;
        assignment.init.assigned_previous.push_back(membert());
      }
      else
      {
        synth_expr(lhs, S_SYMBOL);
        synth_expr(rhs, S_SYMBOL);

        if(assignment.type==G_NONE)
          assignment.type=G_COMBINATORIAL;
        else if(assignment.type!=G_COMBINATORIAL)
        {
          err_location(statement);
          throw "variable is clocked";
        }

        equality_exprt equality(lhs, rhs);
        trans.invar().move_to_operands(equality);
      }
    }
  }
}

/*******************************************************************\

Function: verilog_synthesist::synth_block

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_block(const verilog_blockt &statement)
{
  forall_operands(it, statement)
    synth_statement(static_cast<const verilog_statementt &>(*it));
}

/*******************************************************************\

Function: verilog_synthesist::synth_continuous_assign

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_continuous_assign(
  const verilog_continuous_assignt &module_item,
  transt &trans)
{
  construct=C_OTHER;

  forall_operands(it, module_item)
  {
    if(it->id()!=ID_equal || it->operands().size()!=2)
    {
      err_location(*it);
      str << "unexpected continuous assignment" << std::endl;
      throw 0;
    }

    exprt lhs=it->op0();
    exprt rhs=it->op1();
    
    synth_expr(lhs, S_SYMBOL);
    synth_expr(rhs, S_SYMBOL);

    synth_continuous_assign(lhs, rhs, trans);
  }
}

/*******************************************************************\

Function: verilog_synthesist::synth_continuous_assign

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_continuous_assign(
  exprt &lhs,
  exprt &rhs,
  transt &trans)
{
  if(lhs.id()==ID_concatenation)
  {
    // split it up
    unsigned offset=0;
    
    // do it from right to left
    
    for(exprt::operandst::reverse_iterator
        it=lhs.operands().rbegin();
        it!=lhs.operands().rend();
        it++)
    {
      exprt offset_constant(ID_constant, typet(ID_natural));
      offset_constant.set(ID_value, offset);

      if(it->type().id()==ID_bool)
      {
        exprt bit_extract(ID_extractbit, it->type());
        bit_extract.copy_to_operands(rhs, offset_constant);
        offset++;

        synth_continuous_assign(*it, bit_extract, trans);
      }
      else if(it->type().id()==ID_signedbv ||
              it->type().id()==ID_unsignedbv)
      {
        unsigned width=get_width(it->type());
      
        exprt offset_constant2(ID_constant, typet(ID_natural));
        offset_constant2.set(ID_value, offset+width-1);

        exprt bit_extract(ID_extractbits, it->type());
        bit_extract.copy_to_operands(rhs, offset_constant, offset_constant2);
        
        synth_continuous_assign(*it, bit_extract, trans);
        
        offset+=width;
      }
      else
      {
        err_location(*it);
        str << "continuous assignment to type `" << to_string(it->type())
            << "' not supported";
      }
    }
    
    return;
  }

  // get symbol

  symbolt &symbol=assignment_symbol(lhs);

  assignmentt &assignment=assignments[symbol.name];

  if(assignment.type==G_NONE)
    assignment.type=G_COMBINATORIAL;
  else if(assignment.type!=G_COMBINATORIAL)
  {
    err_location(lhs);
    throw "variable is clocked";
  }

  equality_exprt equality(lhs, rhs);
  trans.invar().move_to_operands(equality);
}

/*******************************************************************\

Function: verilog_synthesist::synth_assign

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_assign(
  const exprt &statement,
  bool blocking)
{
  if(statement.operands().size()!=2)
  {
    err_location(statement);
    throw "assign statement expected to have two operands";
  }

  if(construct!=C_ALWAYS && construct!=C_INITIAL)
  {
    err_location(statement);
    throw "unexpected assign statement";
  }
  
  const exprt &lhs=statement.op0();
  exprt rhs=statement.op1();

  synth_expr(rhs, S_CURRENT);

  // elaborate now?
  if(lhs.type().id()==ID_integer)
  {
    mp_integer i;
    simplify(rhs, ns);

    if(to_integer(rhs, i))
    {
      err_location(rhs);
      throw "synthesis expects constant on rhs";
    }
    
    if(lhs.id()!=ID_symbol)
    {
      err_location(lhs);
      throw "synthesis expects symbol on lhs";
    }
    
    values[to_symbol_expr(lhs).get_identifier()]=i;
  }
  else
    assignment(lhs, rhs, blocking);
}

/*******************************************************************\

Function: verilog_synthesist::synth_assert

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_assert(
  const verilog_assertt &statement)
{
  if(statement.operands().size()!=2)
  {
    err_location(statement);
    throw "assert statement expected to have two operands";
  }

  const irep_idt &identifier=statement.get(ID_identifier);
  symbolt &symbol=context_lookup(identifier);
  
  synth_expr(symbol.value, S_CURRENT);
}

/*******************************************************************\

Function: verilog_synthesist::synth_assert_module_item

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_assert_module_item(
  const verilog_module_itemt &module_item)
{
  if(module_item.operands().size()!=2)
  {
    err_location(module_item);
    throw "assert module_item expected to have two operands";
  }

  const irep_idt &identifier=module_item.get(ID_identifier);
  symbolt &symbol=context_lookup(identifier);
  
  construct=C_OTHER;

  synth_expr(symbol.value, S_SYMBOL);
}

/*******************************************************************\

Function: verilog_synthesist::synth_assume

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_assume(
  const verilog_assumet &statement)
{
  if(statement.operands().size()!=2)
  {
    err_location(statement);
    throw "assume statement expected to have two operands";
  }
  
  exprt condition=statement.condition();

  construct=C_OTHER;

  synth_expr(condition, S_CURRENT);

  // add it as an invariant
  invars.push_back(condition);
}

/*******************************************************************\

Function: verilog_synthesist::synth_assume_module_item

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_assume_module_item(
  const verilog_module_itemt &module_item)
{
  if(module_item.operands().size()!=2)
  {
    err_location(module_item);
    throw "assume module item expected to have two operands";
  }

  exprt condition=module_item.op0();
  synth_expr(condition, S_SYMBOL);

  // add it as an invariant
  invars.push_back(condition);
}

/*******************************************************************\

Function: verilog_synthesist::synth_case_values

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

exprt verilog_synthesist::case_comparison(
  const exprt &case_operand,
  const exprt &pattern)
{
  // we need to take case of ?, x, z in the pattern
  if(pattern.id()=="verilog-constant")
  {
    // we do this using masking!
    unsignedbv_typet new_type;
    new_type.set(ID_width, pattern.type().get(ID_width));
    
    std::string new_pattern_value=
      id2string(pattern.get(ID_value));

    for(unsigned i=0; i<new_pattern_value.size(); i++)
      if(new_pattern_value[i]=='?' ||
         new_pattern_value[i]=='z' ||
         new_pattern_value[i]=='x')
        new_pattern_value[i]='0';
    
    constant_exprt new_pattern(new_type);
    new_pattern.set_value(new_pattern_value);

    std::string new_mask_value=
      id2string(pattern.get(ID_value));

    for(unsigned i=0; i<new_mask_value.size(); i++)
      if(new_mask_value[i]=='?' ||
         new_mask_value[i]=='z' ||
         new_mask_value[i]=='x')
        new_mask_value[i]='0';
      else
        new_mask_value[i]='1';

    constant_exprt new_mask(new_type);
    new_mask.set_value(new_mask_value);
    
    exprt bitand_expr=bitand_exprt(case_operand, new_mask);
    
    if(bitand_expr.type()!=case_operand.type())
      bitand_expr.make_typecast(case_operand.type());
    
    return equality_exprt(bitand_expr, new_pattern);
  }
  
  return equality_exprt(case_operand, pattern);
}

/*******************************************************************\

Function: verilog_synthesist::synth_case_values

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_case_values(
  const exprt &values,
  const exprt &case_operand,
  exprt &guard)
{
  if(values.id()==ID_default)
  {
    guard.make_true();
    return;
  }

  guard.reserve_operands(values.operands().size());

  forall_operands(it, values)
  {
    exprt pattern=*it;
    synth_expr(pattern, S_NONE);
    guard.copy_to_operands(case_comparison(case_operand, pattern));
  }

  gen_or(guard);
}

/*******************************************************************\

Function: verilog_synthesist::synth_case

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_case(
  const verilog_statementt &statement)
{
  if(statement.operands().size()<1)
  {
    err_location(statement);
    throw "case statement expected to have at least one operand";
  }

  exprt case_operand=statement.op0();

  synth_expr(case_operand, S_CURRENT);

  // we convert this to if-then-else

  exprt start;
  exprt *last_if=&start;

  for(unsigned i=1; i<statement.operands().size(); i++)
  {
    const exprt &e=statement.operands()[i];

    if(e.id()!=ID_case_item)
      throw "expected case_item";

    if(e.operands().size()!=2)
    {
      err_location(e);
      throw "case_item expected to have two operands";
    }

    exprt if_statement(ID_if);
    if_statement.reserve_operands(3);
    if_statement.operands().resize(2);

    synth_case_values(
      e.op0(), case_operand, if_statement.op0());

    if_statement.op1()=e.op1();

    last_if->move_to_operands(if_statement);
    last_if=&last_if->operands().back();
  }

  // synthesize the new if-then-else

  if(!start.operands().empty())
    synth_statement(static_cast<verilog_statementt &>(start.op0()));
}

/*******************************************************************\

Function: verilog_synthesist::synth_if

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_if(
  const verilog_ift &statement)
{
  if(statement.operands().size()!=2 &&
     statement.operands().size()!=3)
  {
    err_location(statement);
    throw "if statement expected to have two or three operands";
  }

  exprt if_operand=statement.condition();
  synth_expr(if_operand, S_CURRENT);

  if(if_operand.is_true())
  {
    synth_statement(statement.true_case());
    return;
  }

  // save current value_map pointer
  value_mapt *old_map=value_map;

  // produce new value maps
  value_mapt true_map(*value_map), false_map(*value_map);
  true_map.clear_changed();
  false_map.clear_changed();

  // true case
  {
    value_map=&true_map;
    synth_statement(statement.true_case());
  }
  
  // false case
  if(statement.operands().size()==3)
  {
    value_map=&false_map;
    synth_statement(statement.false_case());
  }

  value_map=old_map;
  merge(if_operand, true_map.current, false_map.current, value_map->current);
  merge(if_operand, true_map.final, false_map.final, value_map->final);
}

/*******************************************************************\

Function: verilog_synthesist::merge

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::merge(
  const exprt &guard,
  const value_mapt::mapt &true_map,
  const value_mapt::mapt &false_map,
  value_mapt::mapt &dest)
{
  // get true_map.changed + false_map.changed

  std::set<irep_idt> changed;

  changed.insert(true_map.changed.begin(), true_map.changed.end());
  changed.insert(false_map.changed.begin(), false_map.changed.end());

  for(std::set<irep_idt>::const_iterator
      it=changed.begin();
      it!=changed.end();
      it++)
  {
    const symbolt &symbol=lookup(*it);

    exprt true_value, false_value;
    current_value(true_map, symbol, false, true_value);
    current_value(false_map, symbol, false, false_value);

    exprt value(ID_if, symbol.type);
    value.operands().resize(3);
    value.op0()=guard;
    value.op1().swap(true_value);
    value.op2().swap(false_value);

    dest.symbol_map[symbol.name].swap(value);
    dest.changed.insert(symbol.name);
  }
}

/*******************************************************************\

Function: verilog_synthesist::synth_event_guard

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_event_guard(
  const verilog_event_guardt &statement)
{
  if(statement.operands().size()!=2)
  {
    err_location(statement);
    throw "event_guard expected to have two operands";
  }

  if(event_guard!=G_NONE)
  {
    err_location(statement);
    throw "event guard already specified";
  }

  const exprt &event_guard_expr=statement.guard();

  bool edge=false;

  // these guards are ORed
  exprt guards; 

  forall_operands(it, event_guard_expr)
    if(it->id()==ID_posedge || it->id()==ID_negedge)
    {
      edge=true;

      if(it->operands().size()!=1)
      {
        err_location(*it);
        throw "pos/negedge expected to have one operand";
      }

      if(it->op0().id()!=ID_symbol)
      {
        err_location(*it);
        throw "pos/negedge expected to have symbol as operand "
              "but got "+it->op0().pretty();
      }

      if(it->op0().type().id()!=ID_bool)
      {
        err_location(*it);
        throw "pos/negedge expected to have Boolean as operand, "
              "but got "+to_string(it->op0().type());
      }

      irep_idt identifier="conf::clock_enable_mode";

      // check context for clock guard

      //std::cout << "DEBUG " << identifier << std::endl;

      if(context.symbols.find(identifier)!=context.symbols.end())
      {
        // found! we make it a guard

        guards.copy_to_operands(it->op0());

        str << "Notice: using clock guard " << identifier << std::endl;
      }
    }

  event_guard=edge?G_CLOCK:G_COMBINATORIAL;

  if(guards.operands().empty())
    synth_statement(statement.body());
  else
  {
    // new guards!
    gen_or(guards);

    value_mapt *old_map=value_map;
    value_mapt true_map(*value_map), false_map(*value_map);
    true_map.clear_changed();
    false_map.clear_changed();

    value_map=&true_map;
    synth_statement(statement.body());

    value_map=old_map;
    merge(guards, true_map.current, false_map.current, value_map->current);
    merge(guards, true_map.final, false_map.final, value_map->final);
  }

  event_guard=G_NONE;
}

/*******************************************************************\

Function: verilog_synthesist::synth_delay

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_delay(
  const verilog_delayt &statement)
{
  if(statement.operands().size()!=2)
  {
    err_location(statement);
    throw "delay expected to have two operands";
  }

  synth_statement(statement.body());
}

/*******************************************************************\

Function: verilog_synthesist::synth_for

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_for(const verilog_fort &statement)
{
  if(statement.operands().size()!=4)
  {
    err_location(statement);
    throw "for expected to have four operands";
  }

  synth_statement(statement.initialization());

  while(true)
  {  
    exprt tmp_guard=statement.condition();
    tmp_guard.make_typecast(bool_typet());
    synth_expr(tmp_guard, S_CURRENT);
    simplify(tmp_guard, ns);
 
    if(!tmp_guard.is_constant())
    {
      err_location(statement.op1());
      str << "synthesis failed to evaluate loop guard: `"
          << expr2verilog(tmp_guard) << "'";
      throw 0;
    }

    if(tmp_guard.is_false()) break;
    
    // copy the body
    verilog_statementt tmp_body=statement.body();
    synth_statement(tmp_body);

    verilog_statementt tmp_inc=statement.inc_statement();
    synth_statement(tmp_inc);
  }
}

/*******************************************************************\

Function: verilog_synthesist::synth_prepostincdec

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_prepostincdec(const verilog_statementt &statement)
{
  if(statement.operands().size()!=1)
  {
    err_location(statement);
    throw statement.id_string()+" expected to have one operand";
  }

  // stupid implementation
  exprt one=
    from_integer(1, statement.op0().type());
  
  verilog_blocking_assignt assignment;
  assignment.lhs()=statement.op0();
  assignment.rhs()=plus_exprt(statement.op0(), one);
  assignment.location()=statement.location();
  synth_statement(assignment);  
}

/*******************************************************************\

Function: verilog_synthesist::synth_while

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_while(
  const verilog_whilet &statement)
{
  if(statement.operands().size()!=2)
  {
    err_location(statement);
    throw "while expected to have two operands";
  }

  while(true)
  {  
    exprt tmp_guard=statement.condition();
    tmp_guard.make_typecast(bool_typet());
    synth_expr(tmp_guard, S_CURRENT);
    simplify(tmp_guard, ns);
 
    if(!tmp_guard.is_constant())
    {
      err_location(statement.op1());
      str << "synthesis failed to evaluate loop guard: `"
          << expr2verilog(tmp_guard) << "'";
      throw 0;
    }

    if(tmp_guard.is_false()) break;

    // copy the body!    
    verilog_statementt tmp_body=statement.body();
    synth_statement(tmp_body);
  }
}

/*******************************************************************\

Function: verilog_synthesist::synth_repeat

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_repeat(
  const verilog_repeatt &statement)
{
  if(statement.operands().size()!=2)
  {
    err_location(statement);
    throw "repeat expected to have two operands";
  }

  err_location(statement);
  throw "cannot synthesize `repeat'";
}

/*******************************************************************\

Function: verilog_synthesist::synth_forever

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_forever(
  const verilog_forevert &statement)
{
  if(statement.operands().size()!=1)
  {
    err_location(statement);
    throw "forever expected to have one operand";
  }

  err_location(statement);
  throw "cannot synthesize `forever'";
}

/*******************************************************************\

Function: verilog_synthesist::synth_task_enable

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_task_enable(
  const verilog_task_enablet &statement)
{
  if(statement.operands().size()!=2)
  {
    err_location(statement);
    throw "task_enable expected to have two operands";
  }

  // this is essentially inlined
  const symbol_exprt &task=to_symbol_expr(statement.task());
  
  const symbolt &symbol=lookup(task.get_identifier());
  
  if(symbol.type.id()!=ID_code)
  {
    err_location(statement);
    throw "task_enable expected task as first operand";
  }
  
  const code_typet &code_type=to_code_type(symbol.type);

  if(code_type.return_type()!=empty_typet())
  {
    err_location(statement);
    throw "task_enable cannot call function";
  }
  
  const code_typet::argumentst &arguments=
    code_type.arguments();

  const exprt::operandst &actuals=
    statement.op1().operands();
    
  if(arguments.size()!=actuals.size())
  {
    err_location(statement);
    throw "wrong number of arguments";
  }
  
  // do assignments to input arguments
  for(unsigned i=0; i<arguments.size(); i++)
  {
    const symbolt &a_symbol=lookup(arguments[i].get_identifier());
    if(arguments[i].get_bool(ID_input))
    {
      verilog_blocking_assignt assignment;
      assignment.lhs()=::symbol_expr(a_symbol);
      assignment.rhs()=actuals[i];
      assignment.location()=statement.location();
      synth_statement(assignment);
    }
  }

  synth_statement(to_verilog_statement(symbol.value));

  // do assignments to output arguments
  for(unsigned i=0; i<arguments.size(); i++)
  {
    const symbolt &a_symbol=lookup(arguments[i].get_identifier());
    if(arguments[i].get_bool(ID_output))
    {
      verilog_blocking_assignt assignment;
      assignment.lhs()=actuals[i];
      assignment.rhs()=::symbol_expr(a_symbol);
      assignment.location()=statement.location();
      synth_statement(assignment);
    }
  }

}

/*******************************************************************\

Function: verilog_synthesist::synth_statement

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_statement(
  const verilog_statementt &statement)
{
  if(statement.id()==ID_block)
    synth_block(to_verilog_block(statement));
  else if(statement.id()==ID_case ||
          statement.id()==ID_casex ||
          statement.id()==ID_casez)
    synth_case(statement);
  else if(statement.id()==ID_blocking_assign)
    synth_assign(statement, true);
  else if(statement.id()==ID_continuous_assign)
  {
    err_location(statement);
    error("synthesis of procedural continuous assignment not supported");
    throw 0;
  }
  else if(statement.id()==ID_assert)
    synth_assert(to_verilog_assert(statement));
  else if(statement.id()==ID_assume)
    synth_assume(to_verilog_assume(statement));
  else if(statement.id()==ID_non_blocking_assign)
    synth_assign(statement, false);
  else if(statement.id()==ID_if)
    synth_if(to_verilog_if(statement));
  else if(statement.id()==ID_event_guard)
    synth_event_guard(to_verilog_event_guard(statement));
  else if(statement.id()==ID_delay)
    synth_delay(to_verilog_delay(statement));
  else if(statement.id()==ID_for)
    synth_for(to_verilog_for(statement));
  else if(statement.id()==ID_while)
    synth_while(to_verilog_while(statement));
  else if(statement.id()==ID_repeat)
    synth_repeat(to_verilog_repeat(statement));
  else if(statement.id()==ID_forever)
    synth_forever(to_verilog_forever(statement));
  else if(statement.id()==ID_task_enable)
    synth_task_enable(to_verilog_task_enable(statement));
  else if(statement.id()==ID_preincrement ||
          statement.id()==ID_predecrement ||
          statement.id()==ID_postincrement ||
          statement.id()==ID_postdecrement)
    synth_prepostincdec(statement);
  else if(statement.id()==ID_skip)
  {
    // do nothing
  }
  else
  {
    err_location(statement);
    str << "unexpected statement during synthesis: "
        << statement.id();
    throw 0;
  }
}

/*******************************************************************\

Function: verilog_synthesist::synth_module_item

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_module_item(
  const verilog_module_itemt &module_item,
  transt &trans)
{
  if(module_item.id()=="specify")
  {
  }
  else if(module_item.id()==ID_decl)
    synth_decl(to_verilog_decl(module_item), trans);
  else if(module_item.id()=="parameter_decl")
  {
  }
  else if(module_item.id()==ID_always)
    synth_always(to_verilog_always(module_item));
  else if(module_item.id()==ID_initial)
    synth_initial(to_verilog_initial(module_item));
  else if(module_item.id()==ID_continuous_assign)
    synth_continuous_assign(to_verilog_continuous_assign(module_item), trans);
  else if(module_item.id()==ID_inst)
    synth_inst(to_verilog_inst(module_item), trans);
  else if(module_item.id()=="inst-builtin")
    synth_inst_builtin(module_item, trans);
  else if(module_item.id()==ID_generate_block)
  {
    err_location(module_item);
    str << "unexpected generate module item during synthesis" << std::endl;
    throw 0;
  }
  else if(module_item.id()==ID_assert)
    synth_assert_module_item(module_item);
  else if(module_item.id()==ID_assume)
    synth_assume_module_item(module_item);
  else if(module_item.id()==ID_task)
  {
    // ignore for now
  }
  else
  {
    err_location(module_item);
    str << "unexpected module item during synthesis:"
        << module_item.id();
    throw 0;
  }
}

/*******************************************************************\

Function: verilog_synthesist::symbol_expr

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::symbol_expr(
  const symbolt &symbol,
  bool next,
  exprt &dest)
{
  dest=exprt(next?ID_next_symbol:ID_symbol, symbol.type);
  dest.set(ID_identifier, symbol.name);
}

/*******************************************************************\

Function: verilog_synthesist::synth_assignments

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_assignments(
  const symbolt &symbol,
  bool next,
  exprt &new_value,
  exprt &constraints)
{
  if(new_value.is_nil())
    symbol_expr(symbol, false, new_value);
  
  // see if wire is used to define itself
  if(!symbol.is_statevar)
    post_process_wire(symbol.name, new_value);

  equality_exprt equality_expr;

  symbol_expr(symbol, next, equality_expr.lhs());
  equality_expr.rhs().swap(new_value);

  constraints.move_to_operands(equality_expr);
}

/*******************************************************************\

Function: verilog_synthesist::synth_assignments

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::synth_assignments(transt &trans)
{
  for(local_symbolst::const_iterator
      it=local_symbols.begin();
      it!=local_symbols.end();
      it++)
  {
    symbolt &symbol=context_lookup(*it);

    if(symbol.is_statevar && !symbol.is_macro)
    {
      assignmentt &assignment=assignments[symbol.name];

      if(assignment.type==G_COMBINATORIAL)
      {
        str << "Making " << symbol.name << " a wire" << std::endl;
        symbol.is_statevar=false;
      }

      if(symbol.is_statevar)
      {
        // only state variables can be initialized

        if(!assignment.init.value.is_nil())
          synth_assignments(symbol, false,
                            assignment.init.value,
                            trans.op1());

        synth_assignments(symbol, true,
                          assignment.next.value,
                          trans.op2());
      }
      else
      {
        synth_assignments(symbol, false,
                          assignment.next.value,
                          trans.invar());
      }
    }
  }

  // post-process initial state predicate to get rid
  // of unnecessary nondet_symbols

  post_process_initial(trans.op1());
}

/*******************************************************************\

Function: verilog_synthesist::current_value

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::current_value(
  const value_mapt::mapt &map,
  const symbolt &symbol,
  bool use_previous_assignments,
  exprt &dest)
{
  if(!symbol.is_statevar)
  {
    if(use_previous_assignments)
    {
      // see if we have a previous assignment
      const assignmentt &assignment=assignments[symbol.name];
      const exprt &value=
        (construct==C_INITIAL)?assignment.init.value:assignment.next.value;

      if(value.is_not_nil())
      {
        dest=value;
        return; // done
      }
    }

    symbol_expr(symbol, false, dest);
  }
  else // latch
  {
    value_mapt::mapt::symbol_mapt::const_iterator it=
      map.symbol_map.find(symbol.name);
      
    if(it==map.symbol_map.end())
    {
      if(use_previous_assignments)
      {
        // see if we have a previous assignment
        const assignmentt &assignment=assignments[symbol.name];
        const exprt &value=
          (construct==C_INITIAL)?assignment.init.value:assignment.next.value;

        if(value.is_not_nil())
        {
          dest=value;
          return; // done
        }
      }

      if(construct==C_ALWAYS)
        symbol_expr(symbol, false, dest);
      else
      {
        // make it some random value
        dest=exprt(ID_nondet_symbol, symbol.type);
        dest.set(ID_identifier, symbol.name);
        dest.set("initial_value", true);
      }
    }
    else
      dest=it->second;
  }
}

/*******************************************************************\

Function: subexpressions

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void subexpressions(const exprt &expr, std::multiset<exprt> &counters)
{
  counters.insert(expr);

  forall_operands(it, expr)
    subexpressions(*it, counters);
}

/*******************************************************************\

Function: verilog_synthesist::post_process_initial

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::post_process_initial(exprt &constraints)
{
  // look for unused non-determinism constraints
  
  std::multiset<exprt> counters;

  forall_operands(it, constraints)
    subexpressions(*it, counters);

  Forall_operands(it, constraints)
  {
    if(it->id()==ID_equal && it->operands().size()==2)
    {
      exprt &lhs=it->op0(),
            &rhs=it->op1();

      #if 0
      if(lhs.id()==ID_symbol && 
         rhs.id()=="nondet_symbol" &&
         lhs.get(ID_identifier)==rhs.get(ID_identifier))
      #else
      if(lhs.id()==ID_symbol && 
         rhs.id()=="nondet_symbol")
      #endif
      {
        if(counters.count(rhs)==1)
        {
          // not used elsewhere
          it->make_true();
        }
      }
    }
  }
}

/*******************************************************************\

Function: verilog_synthesist::post_process_wire

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::post_process_wire(
  const irep_idt &identifier,
  exprt &expr)
{
  // look if wire is used to define wire itself
  
  Forall_operands(it, expr)
    post_process_wire(identifier, *it);

  if(expr.id()==ID_symbol && 
     expr.get(ID_identifier)==identifier)
  {
    expr.id(ID_nondet_symbol);
  }
}

/*******************************************************************\

Function: verilog_synthesist::convert_module_items

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::convert_module_items(symbolt &symbol)
{
  assert(symbol.value.id()==ID_verilog_module);

  transt trans;

  // clean up
  assignments.clear();
  invars.clear();

  // find out about symbols of this module

  forall_symbol_module_map(it, context.symbol_module_map, module)
    local_symbols.insert(it->second);

  // do the module items

  forall_operands(it, symbol.value)
    synth_module_item(
      static_cast<const verilog_module_itemt &>(*it), trans);

  synth_assignments(trans);
  
  for(std::list<exprt>::const_iterator
      it=invars.begin();
      it!=invars.end();
      it++)
    trans.invar().copy_to_operands(*it);

  gen_and(trans.invar()); // general
  gen_and(trans.init());  // initial
  gen_and(trans.trans()); // transition relation

  symbol.value=trans;
}

/*******************************************************************\

Function: verilog_synthesist::typecheck

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void verilog_synthesist::typecheck()
{
  symbolt &symbol=context_lookup(module);
  if(symbol.value.id()==ID_trans) return; // done already
  convert_module_items(symbol);
}

/*******************************************************************\

Function: verilog_synthesis

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool verilog_synthesis(
  contextt &context,
  const irep_idt &module,
  message_handlert &message_handler,
  const optionst &options)
{
  verilog_synthesist verilog_synthesis(
    context, module, options, message_handler);
  return verilog_synthesis.typecheck_main();
}
